import { BaseEvent } from '@/application/base/baseEvent'
import { SocketEvent } from '@/application/enums'
import type { MapEventTileWithTeleport } from '@/application/types'
import MapManager from '@/managers/mapManager'
import MapCharacter from '@/models/mapCharacter'
import MapEventTileRepository from '@/repositories/mapEventTileRepository'
import CharacterService from '@/services/characterMoveService'
import TeleportService from '@/services/characterTeleportService'

export default class CharacterMove extends BaseEvent {
  private readonly characterService = CharacterService
  private readonly STEP_DELAY = 150 // Milliseconds between each tile movement
  private readonly THROTTLE_DELAY = 230 // Minimum time between movement requests
  private movementTimeout: NodeJS.Timeout | null = null

  public listen(): void {
    this.socket.on(SocketEvent.MAP_CHARACTER_MOVE, this.handleEvent.bind(this))
  }

  private async handleEvent({ positionX, positionY }: { positionX: number; positionY: number }): Promise<void> {
    try {
      const mapCharacter = MapManager.getCharacterById(this.socket.characterId!)
      if (!mapCharacter?.getCharacter()) {
        this.logger.error('map:character:move error: Character not found or not initialized')
        return
      }

      // Cancel any ongoing movement
      this.cancelCurrentMovement(mapCharacter)

      // Throttle rapid movement requests
      if (this.isThrottled(`movement_${this.socket.characterId}`, this.THROTTLE_DELAY)) {
        return
      }

      // Calculate path to target position
      const path = await this.characterService.calculatePath(mapCharacter.character, Math.floor(positionX), Math.floor(positionY))

      if (!path?.length) {
        this.io.in(mapCharacter.character.map.id).emit(SocketEvent.MAP_CHARACTER_MOVEERROR, 'No valid path found')
        return
      }

      // Start new movement
      mapCharacter.isMoving = true
      mapCharacter.currentPath = path
      await this.moveAlongPath(mapCharacter)
    } catch (error: any) {
      this.logger.error('map:character:move error: ' + error.message)
    }
  }

  private cancelCurrentMovement(mapCharacter: MapCharacter): void {
    if (mapCharacter.isMoving) {
      mapCharacter.isMoving = false
      mapCharacter.currentPath = null
    }
  }

  private async moveAlongPath(mapCharacter: MapCharacter): Promise<void> {
    const character = mapCharacter.getCharacter()
    const path = mapCharacter.currentPath

    if (!path?.length) return

    try {
      for (let i = 0; i < path.length - 1; i++) {
        if (!mapCharacter.isMoving || mapCharacter.currentPath !== path) {
          return
        }

        const [currentTile, nextTile] = [path[i], path[i + 1]]

        if (!currentTile || !nextTile) {
          this.logger.error('Invalid movement step detected')
          break
        }

        // Validate step
        if (!this.isValidStep(currentTile, nextTile)) {
          this.logger.error('Invalid movement step detected')
          break
        }

        // Update character rotation
        character.setRotation(CharacterService.calculateRotation(currentTile.positionX, currentTile.positionY, nextTile.positionX, nextTile.positionY))

        // Check for map events at the next tile
        const mapEventTile = await this.checkMapEvents(character, nextTile)
        if (mapEventTile) {
          if (mapEventTile.type === 'BLOCK') break
          if (mapEventTile.type === 'TELEPORT' && mapEventTile.teleport) {
            await this.handleTeleportMapEventTile(mapEventTile as MapEventTileWithTeleport)
            return
          }
        }

        // Move character to next tile
        character.setPositionX(nextTile.positionX).setPositionY(nextTile.positionY)

        // Broadcast movement
        this.broadcastMovement(character, true)

        // Apply movement delay between steps
        await new Promise((resolve) => setTimeout(resolve, this.STEP_DELAY))
        if (i < path.length - 2) {
        }
      }
    } finally {
      if (mapCharacter.isMoving && mapCharacter.currentPath === path) {
        this.finalizeMovement(mapCharacter)
      }
    }
  }

  private isValidStep(current: { positionX: number; positionY: number }, next: { positionX: number; positionY: number }): boolean {
    return Math.abs(next.positionX - current.positionX) <= 1 && Math.abs(next.positionY - current.positionY) <= 1
  }

  private async checkMapEvents(character: any, nextTile: { positionX: number; positionY: number }) {
    const mapEventTileRepository = new MapEventTileRepository()
    return mapEventTileRepository.getEventTileByMapIdAndPosition(character.getMap().getId(), Math.floor(nextTile.positionX), Math.floor(nextTile.positionY))
  }

  private async handleTeleportMapEventTile(mapEventTile: MapEventTileWithTeleport): Promise<void> {
    const teleport = mapEventTile.getTeleport()
    if (teleport) {
      await TeleportService.teleportCharacter(this.socket.characterId!, {
        targetMapId: teleport.getToMap().getId(),
        targetX: teleport.getToPositionX(),
        targetY: teleport.getToPositionY(),
        rotation: teleport.getToRotation()
      })
    }
  }

  private broadcastMovement(character: any, isMoving: boolean): void {
    this.io.in(character.map.id).emit(SocketEvent.MAP_CHARACTER_MOVE, {
      characterId: character.id,
      positionX: character.getPositionX(),
      positionY: character.getPositionY(),
      rotation: character.getRotation(),
      isMoving
    })
  }

  private finalizeMovement(mapCharacter: MapCharacter): void {
    // Clear any existing timeout
    if (this.movementTimeout) {
      clearTimeout(this.movementTimeout)
    }

    // Set new timeout
    this.movementTimeout = setTimeout(() => {
      // Only finalize if there are no pending movements
      if (mapCharacter.currentPath === null) {
        mapCharacter.isMoving = false
        this.broadcastMovement(mapCharacter.character, false)
      }
    }, this.STEP_DELAY)

    mapCharacter.currentPath = null
  }
}
